/*
 * LC7461M8100 simulator
 *
 * A. Hojka
 * November 2018
 * */

#include <stm32l0xx_ll_gpio.h>
#include "stm32l0xx_ll_cortex.h"
#include "stm32l0xx_ll_tim.h"
#include "stm32l0xx.h"
#include "main_utils.h"
#include "sanyo_ir.h"
#include "main.h"

//------------------------------------------------------
volatile uint32_t			AppFlagReg;					//application state register

#define APP_REQ_TOSTOP		(uint32_t) (0x01 << 0)		//if there is no activity go to STOP mode
#define APP_DEF_VARS		(uint32_t) (0x01 << 1)		//zero all variables - resuming from stop
#define APP_SCAN_KEY_RDY	(uint32_t) (0x01 << 2)		//current scan completed
#define APP_SCAN_INVALID	(uint32_t) (0x01 << 3)		//invalid multi press
#define APP_KEY_PRESSED		(uint32_t) (0x01 << 4)		//we have key down - may be invalid - doesn't matter
#define APP_KEY_HELD		(uint32_t) (0x01 << 5)		//
#define APP_TXRQ			(uint32_t) (0x01 << 6)		//set in main loop, cleared by ISR -> 0 - idle/ready,  1-request to send IR frame is pending
#define APP_USEBUF_B		(uint32_t) (0x01 << 7)		//1 - put frame to B when REQ, 0 - use buffer A
#define APP_TX_MODSTATE		(uint32_t) (0x01 << 8)		//frame always starts from 1 so subsequent states are negation

//------------------------------------------------------

volatile uint8_t KeyDataFromISR;		//written by ISR along with APP_KEY_PRESSED flag
volatile uint8_t IR_FrameBuf_A[89];		//2 buffers for IR frame to send, encoding in main loop, key scanning and transmission in ISR
volatile uint8_t IR_FrameBuf_B[89];		//2(SOF) + 2 x 26 + 2 x 16 + 1 + 1 + 1 = 89B

//-------------------------------------------------------

int main(void)
{
	GPIO_cfg();			//setup peripherals after POR
	CLOCK_cfg();
	PWR_cfg();
	TIM21_cfg();
	NVIC_cfg();			//the only one ISR is SysTick Handler exception
	SYSTICK_cfg();
	EXTI_cfg();			//wake up on any keyboard activity

	while(1){		// --- main loop ---
		STOPModeEnter();//prepare peripherals to STOP
						//magic sequence to enter STOP mode
		__SEV();		//this is in case of any remaining event preventing successful STOP 
		__WFE();		//"If the Event Register is set, WFE clears it and returns immediately."
						//still here...
		__WFE();		//now enter STOP mode

			/*
			 * long live the battery!
			 * */
					//we are back
		AppFlagReg = APP_DEF_VARS;	//first ISR will reset variables and prepare reliable key scan
		STOPModeExit();				//activate all required peripherals including interrupt
		LL_LPM_EnableSleep();		//sleep during awaiting for ISR, no sleep on exit

		do{
			__WFI();				//wait for ISR / enter SLEEP mode

			if(AppFlagReg & APP_KEY_PRESSED){			//set only once
				AppFlagReg&=~APP_KEY_PRESSED;

				if(!(AppFlagReg & APP_TXRQ)){			//is there any space ?

					if(AppFlagReg & APP_USEBUF_B){		//---encode IR fame and put to free buffer
						IREncodeFrame((uint8_t *)&IR_FrameBuf_B[0], RC550_CUSTOM_CODE, KeyDataFromISR);
					}else{
						IREncodeFrame((uint8_t *)&IR_FrameBuf_A[0], RC550_CUSTOM_CODE, KeyDataFromISR);
					}
					AppFlagReg|=APP_TXRQ;				//transmission request
				}
			}else if(AppFlagReg & APP_KEY_HELD){			//set periodically
				AppFlagReg&=~APP_KEY_HELD;

				if(!(AppFlagReg & APP_TXRQ)){			//is there any space ?
					if(AppFlagReg & APP_USEBUF_B){		//encode repeat notification
						IREncodeRepFrame((uint8_t *)&IR_FrameBuf_B[0]);
					}else{
						IREncodeRepFrame((uint8_t *)&IR_FrameBuf_A[0]);
					}
					AppFlagReg|=APP_TXRQ;
				}
			}

		}while( !(AppFlagReg & APP_REQ_TOSTOP) );		//do we have long period of inactivity ?
														//go to STOP again...
	}//main loop
}//main




//-------------------------------------------------------
// 			SysTick ISR
//-------------------------------------------------------

uint8_t 	KeyScanCtr;										//counter for key scan (0-31) - gives KeyData directly
uint8_t 	KeyData_Current, KeyData_Previous;				//KeyDN code
uint8_t		KeyInactivityCtr;								//time to sleep
uint8_t		KeyDnCtr;										//hold or not ?

#define 	KEYSCN_KEY_DN		(uint8_t) (0x80)			//if we have valid KeyDN --> KeyData_Current | KEYSCN_KEY_DN


//LL_TIM_OCMODE_PWM1 -> (TIM_CCMR1_OC1M_2 | TIM_CCMR1_OC1M_1)
#define 	PWM_ON()	do{SET_BIT(TIM21->CCMR1, (TIM_CCMR1_OC1M_2 | TIM_CCMR1_OC1M_1) << 8U); TIM21->CNT = 0;}while(0)
//LL_TIM_OCMODE_INACTIVE -> (TIM_CCMR1_OC1M_1)
#define 	PWM_OFF()	do{CLEAR_BIT(TIM21->CCMR1, TIM_CCMR1_OC1M_2 << 8U);}while(0)


void SysTick_Handler(void)
{
static uint8_t *pIRdata = NULL;

	if(AppFlagReg & APP_DEF_VARS){		//first ISR after wake up is idle, zero variables and select first row to scan
		AppFlagReg&=~APP_DEF_VARS;

		KeyPrepare();					//set column before 1st scan

		KeyInactivityCtr 	= 0;
		KeyData_Previous 	= 0;
		KeyData_Current 	= 0;
		KeyScanCtr 			= 0;
		KeyDnCtr			= 0;
		pIRdata 			= NULL;

		AppFlagReg&=~(APP_TXRQ | APP_USEBUF_B | APP_TX_MODSTATE | APP_SCAN_KEY_RDY | APP_SCAN_INVALID);
		return;
	}

	//-------------------------------------------------------
	//							---IR---
	//-------------------------------------------------------
	if(pIRdata != NULL){			//transmission in progress, so continue

		if(--(*pIRdata) == 0){		//count down time of current IR modulation state
			pIRdata++;				//point to next timing

			if(*pIRdata == 0){						//end of frame
				pIRdata = NULL;
			}else{									//next state - set PWM accordingly
				AppFlagReg^=APP_TX_MODSTATE;

				if(AppFlagReg & APP_TX_MODSTATE){	//modulate IR accordingly
					PWM_ON();
					DGB_PIN_ON();
				}else{
					PWM_OFF();
					DGB_PIN_OFF();
				}
			}
		}
	}

	if(pIRdata == NULL){								//no transmission at this moment or just finished ?
		if(AppFlagReg & APP_TXRQ){						//transmission request
			if(AppFlagReg & APP_USEBUF_B){					//where we have data ?
				pIRdata = (uint8_t *)&IR_FrameBuf_B[0];		//this will start transmission
			}else{
				pIRdata = (uint8_t *)&IR_FrameBuf_A[0];
			}

		PWM_ON();						//start IR
		DGB_PIN_ON();

		AppFlagReg|=APP_TX_MODSTATE;	//IR is ON
		AppFlagReg^=APP_USEBUF_B;		//for next frame use different buffer
		AppFlagReg&=~APP_TXRQ;			//second buffer is ready for data
		}								//else{PWM_OFF();}
	}

	//-------------------------------------------------------
	//				---keyboard service routine---
	//-------------------------------------------------------
	KeyScan();

	if(AppFlagReg & APP_SCAN_KEY_RDY){			//entire keyboard scanned

		if( !(KeyData_Current & KEYSCN_KEY_DN) ){
			if(++KeyInactivityCtr > 111){		//4.5ms x 111 = 0.5sec
				AppFlagReg|=APP_REQ_TOSTOP;		//go to sleep
				return;
			}
		}

		if( (KeyData_Current & KEYSCN_KEY_DN) && !(AppFlagReg & APP_SCAN_INVALID) ){	//valid key

			if(KeyData_Previous == KeyData_Current){									//the same as previously
				KeyDnCtr++;
				if(KeyDnCtr == 2){												//2 * 4.5ms = 9ms - two consecutive readings are the same
					KeyDataFromISR = KeyData_Current & ~KEYSCN_KEY_DN;			//key code
					AppFlagReg|=APP_KEY_PRESSED;								//key press info only once
				}else if(KeyDnCtr == 20){
					AppFlagReg|=APP_KEY_HELD;									//repeat hold info
					KeyDnCtr = 2;
				}

			}else{												//not the same data - key debouncing
				KeyDnCtr = 0;
			}

			KeyInactivityCtr = 0;		//stay active
		}else{							//invalid scan or no data, clear all
			KeyDnCtr = 0;				//clear counter and saved data
			KeyData_Previous = 0;
			AppFlagReg&=~(APP_KEY_HELD | APP_KEY_PRESSED);
		}

		AppFlagReg&=~(APP_SCAN_KEY_RDY | APP_SCAN_INVALID);	//clear flags
		KeyData_Previous = KeyData_Current;					//save for comparison
		KeyData_Current = 0;								//clear before next scan
	}//key scan is ready

}//systick handler






/*
Multi-press of key No. 20 and one of key No. 21, 22, 23 may be done, with no priority given in key entry.
When the two keys are kept pressed, a series of pulses will be output according to each key entry.

If multi-press of keys which are not allowed multi-press is done, no output will be delivered.
 */

/*select coulmn before 1st readout*/
void KeyPrepare(void)
{
		LL_GPIO_SetPinOutputType(GPIOA, PIN_Ko0toKo7_ALL, LL_GPIO_OUTPUT_OPENDRAIN);		//open drain works only for low state
		LL_GPIO_SetOutputPin(GPIOA, PIN_Ko0toKo7_ALL);										//all output to high state = high Z
		LL_GPIO_SetPinOutputType(GPIOA, PIN_Ko_0, LL_GPIO_OUTPUT_PUSHPULL);					//push-pull only for selected column
}


/*check 4 buttons and switch to next column, until all 32 buttons checked*/
void KeyScan(void)
{
uint32_t 	ultmp;
uint32_t	keyEntryPort;

	keyEntryPort = LL_GPIO_ReadInputPort(GPIOB);										//timing Pin set in previous tick, read port

	do{																					//do it for 4 possible keys
		if(keyEntryPort & (PIN_Ki_0 << (KeyScanCtr % 4)) ){								//is key pressed ?
			ultmp = KeyScanCtr;															//remeber which one

			if( !(AppFlagReg & APP_SCAN_INVALID) ){									//is scan valid ?

				if(KeyData_Current & KEYSCN_KEY_DN){								//there was a key pressed before

					if( (KeyData_Current == (20 | KEYSCN_KEY_DN)) && (ultmp == 21 || ultmp == 22 || ultmp == 23) ){
						KeyData_Current = (ultmp + 32) | KEYSCN_KEY_DN;				//KEYFLG_KEY_DN is set
					}else{
						AppFlagReg|=APP_SCAN_INVALID;								//invalid multi press - discard current scan
					}

				}else{
					KeyData_Current = ultmp | KEYSCN_KEY_DN;						//regular sequence - first time in scan key is DN
				}
			}

		}//key pressed

		KeyScanCtr++;

	}while(KeyScanCtr % 4 != 0);											//until next column needs to be selected

	if(KeyScanCtr == 32){
		KeyScanCtr = 0;
		AppFlagReg|=APP_SCAN_KEY_RDY;								//keyboard scanned
	}

	LL_GPIO_SetPinOutputType(GPIOA, PIN_Ko0toKo7_ALL, LL_GPIO_OUTPUT_OPENDRAIN);
	LL_GPIO_SetPinOutputType(GPIOA, PIN_Ko_0 << (KeyScanCtr >> 2), LL_GPIO_OUTPUT_PUSHPULL); 				//select new scan line (timing PIN) - KeyScanCtr/4 = 0...7
} //KeyScan







